#
# camera.py
#
#
# Definition of camera class
#
# 1.8.2005 Mazany Ondrej - email:mazany.o@seznam.cz
#

# importing needed modules
import Blender
from Blender import *
from Blender.Scene import Render
from Blender.Mathutils import *
from math import * 
# importing mm function for multiplying matrices
from mymatrix import * 

class CalibCam:
    "A CalibCam class which is used by animator project for camera configuration in Blender"
    def __init__(self, camera = None): # new instance of Camera class
        """
        Inicializes a Camera class, no rotation, position in origin, tga format, 1:1 ratio, 800x600 pixels
        """
        if (camera is None):
            self.name = ""                       # name of object in Blender scene
            self.rotation = [ 0.0, 0.0, 0.0 ]    # rotation around x, y, z axis in radians
            self.location = [ 0.0, 0.0, 0.0 ]    # location in scene x, y, z coords
            self.lens = 16.0                       # lens value of camera
            self.aspx = 1                        # acpect ratio x
            self.aspy = 1                        # x : y - ascpect ratio
            self.sizex = 800                     # output image size in x axis
            self.sizey = 600                     # output image size in y axis        
            self.format = "tga"                  # output image format
            self.ob = None                       # blender object
            self.cam = None
        else:
            self.ob = camera
            self.cam = self.ob.getData()
            self.name = self.ob.getName()        
            self.location = [ self.ob.LocX, self.ob.LocY, self.ob.LocZ ]    
            self.rotation = [ self.ob.RotX, self.ob.RotY, self.ob.RotZ ]    
            self.lens = self.cam.getLens()                     
            cur =  Blender.Scene.getCurrent()
            context = cur.getRenderingContext()            
            self.aspx = context.aspectRatioX()                        
            self.aspy = context.aspectRatioY()                       
            self.sizex = context.imageSizeX()                    
            self.sizey = context.imageSizeY()                     
            self.format = "tga"       
                       
    def setLocRot(self): # sets the location and rotation
        "Sets the camera location and rotation"
        if self.ob <> None:
            self.ob.setLocation( self.location[0], self.location[1], self.location[2]) # sets the location of blender object
            self.ob.RotX = self.rotation[0]            # sets the rotation around x axis of blender object
            self.ob.RotY = self.rotation[1]            # sets the rotation around y axis of blender object
            self.ob.RotZ = self.rotation[2]            # sets the rotation around z axis of blender object
            
    def setUp(self): # sets the camera up and add it to a scene if it doesn't exist
        "Sets the camera up in the Blender. If it doesn't exists, creates a new one."
        try:      
            cur =  Blender.Scene.getCurrent()       # get the current scene
            if self.ob == None: # if no blender object is associated with this camera, we need to find it or create new
                # looking up for camera in the scene
                try:   
                    self.ob = Blender.Object.Get( self.name)    # try to get object with same name   
                # if no such object is found create a new one
                except:      
                    cam = Blender.Camera.New( 'persp', self.name)          # create new perspective camera 
                    self.ob = Blender.Object.New( 'Camera', self.name)     # make camera object
                    self.ob.link( cam)                                    # link camera data with this object
                    cur.link( self.ob)                                    # link object into scene
            # Get the Camera datablock
            self.cam = self.ob.getData()
            # now set the camera parameters up
            self.setLocRot()                                 # set the location and rotation            
            self.cam.setLens( self.lens)                 # set the camera lens
            self.cam.setClipStart( 0.1)                  # set the minimal distance for the closest object to be rendered
            self.cam.setClipEnd( 100)                    # set the maximal distance for the farest object to be rendered
            cur.setCurrentCamera( self.ob)              # activate the camera
            # now set the renderer parameters up
            context = cur.getRenderingContext()         # get the rendering context for current scene
            context.aspectRatioX( self.aspx)            # set the aspect ratio X
            context.aspectRatioY( self.aspy)            # set the aspect ratio Y
            context.imageSizeX( self.sizex)             # set the output format width
            context.imageSizeY( self.sizey)             # set the output format height
            # now set the output format
            if self.format == "jpeg":
                context.setImageType(Render.JPEG90)
            elif self.format == "tga":
                context.setImageType(Render.TARGA)
            elif self.format == "rawtga":
                context.setImageType(Render.RAWTARGA)
            elif self.format == "aviraw":
                context.setImageType(Render.AVIRAW)
            elif self.format == "avijpeg":
                context.setImageType(Render.AVIJPEG)
            elif self.format == "avicodec":
                context.setImageType(Render.AVICODEC)
            elif self.format == "quicktime":
                context.setImageType(Render.QUICKTIME)
            elif self.format == "png":
                context.setImageType(Render.PNG)
            elif self.format == "bmp":
                context.setImageType(Render.BMP)
            elif self.format == "hamx":
                context.setImageType(Render.HAMX)
            elif self.format == "iris":
                context.setImageType(Render.IRIS)
            elif self.format == "irisz":
                context.setImageType(Render.IRIZ)
            elif self.format == "ftype":
                context.setImageType(Render.FTYPE)        
            print "Camera ",self.name," set up succesfuly."
        except:
            print "!! Problem setting ", self.name," camera !!"
    def getK(self):
        "Returns camera calibration matrix with offsets."
        if (self.sizex*self.aspx > self.sizey*self.aspy):
            kv=self.sizex/2.0
            ku=kv*self.aspx/self.aspy
        else:
            ku=self.sizey/2.0
            kv=ku*self.aspy/self.aspx           
        # principal point offset (image origin)
        u0=float(self.sizey)/2.0
        v0=float(self.sizex)/2.0
        # KALIBRATION MATRIX
        f=float(self.lens)/16.0 # this is a konstant we need to divide when transforming from blender lens to focal length
        # THE OLD
        K=Matrix(     [ -f*ku,      0.0,       u0],
                [   0.0,    f*kv,        v0],
                [   0.0,      0.0,      1.0])
        # [0,0] is top left corner, cameras see in negative z
        #K=Matrix(K)
        print "Camera K: ",K
        return K
    def getR(self):
        "Returns matrix which transform world coords into camera coords. It is defined by camera rotation and position"        
        # angles to camera systems are in oposite direction
        sinx=sin((self.rotation[0]))
        cosx=cos((self.rotation[0]))
        siny=sin((self.rotation[1]))
        cosy=cos((self.rotation[1]))
        sinz=sin((self.rotation[2]))
        cosz=cos((self.rotation[2]))
        print "Camera rotation in world: ",self.rotation
        # rotation around z axis
        RZ=Matrix(    [ cosz,   -sinz,   0.0],  
                [ sinz,   cosz,   0.0],
                [    0.0,      0.0,  1.0])
        # rotation around y axis
        RY=Matrix(    [ cosy,   0.0,  siny],
                [    0,   1.0,      0],
                [ -siny,   0,   cosy])
        # rotation around x axis
        RX=Matrix(    [    1.0,   0,       0],
                [    0,   cosx,  -sinx],
                [    0,   sinx,  cosx])
        # ROTATION MATRIX
        # ROT=RX*RY*RZ mm - multiply matrix 
        # MATRIX TRANSFORMING TO CV CAMERA SYSTEM
        T = Matrix(   [ 0.0,  1.0,    0.0],
                [ 1.0,  0.0,    0.0],
                [ 0.0,  0.0,   -1.0])
        # THIS IS HOW THE ROTATIONS IS DEFINED IN BLENDER
        # IT IS THE ROTATION OF CAMERA IN WORLDS COORDS
        
        ROT=RZ*RY*RX
        # WE NEED TO COMPUTE THE WORLD COORDS TO CAMERA ROTATION - inverse rotation
       
        iROT=ROT.invert()
        # AND FINALY THE ROTATION FOR TRANSFORMING THE AXES
        
        R=T*iROT
        #ROT=mm(T,RX)
        #ROT=mm(ROT,RY) 
        #ROT=mm(ROT,RZ)
        # PROJECTIVE ORTHOGONAL MATRIX
        print "Camera R: ",R
        return R
    def getP(self):
        """
        Returns the projective matrix which transform world coords into image coords, where the image origin is upper left corner and v+ direction is down
        Final transformation from worlds x,y,z
        ^ y                         (0,0)   +---------> v
        |                                   |
        |                                   |
        +------> x     =====>               |
         \                                  V u
          \
           V z                  note that camera in blender captures in negative z
        """
        # P=K*R*[eye(3),-C]
        EC=Matrix( [ 1.0,  0,  0, -self.location[0]    ],
            [   0,  1.0,  0, -self.location[1]    ],
            [   0,  0,  1.0, -self.location[2]    ])
        # this  is a hack for correct behaviour of mathutils
        #ECtransposed=Matrix ([1.0,0.0,0.0],[0.0,1.0,0.0],[0.0,0.0,0.1],[-self.location[0],-self.location[1],-self.location[2]])   
        P=mm(self.getK(),self.getR())
        P=mm(P,EC)
        #P=Matrix(P)
        # This is dangerous becouse Blenders.Mathutils can t multply regular matrices 
        # It doesnt work
        #P=self.getK()*self.getR()*EC
        #P=ECtransposed*Matrix(self.getR()).transpose()*Matrix(self.getK).transpose()
        #P.transpose()        
        # Instead using own multiply function mm(A,B)
        #P=mm(self.getK(),self.getR())
        #P=mm(P,EC)
        print "Returned matrix:"
        print P
        return P
    def update(self):
        "Updates camera settings from Blenders camera"
        