import numpy
from scipy import *

import Blender
from Blender import *
from Blender.Mathutils import *
from rahat.glrender import *
from rahat.bfunc import *
import re

# Class Calibrated Image
class CalibIm:
    def __init__(self,imageFile=None,matrixFile=None):
        """
        def __init__(self,imageFile,matrixFile): loads an image and projection matrix
        This inicializes CalibIm object. Calibrated Image. It is a image associated
        with the Projection matrix.
        imageFile is the file with image
        matrixFile is the file with projection matrix in ASCII format like this:
        x.x,x.x,x.x,-x.x;
        x.x,-x.x,x.x,-x.x;
        x.x,x.x,x.x,x.x;
        0.0,0.0,0.0,1.0; this row is not needed
        """
        if imageFile is None:
            self.img = None
            self.P = None
            self.bitmap = None
        else:
            self.img = Image.Load(imageFile)        
            self.loadMatrix(matrixFile)
            self.toBitmap()
    def loadMatrix(self,matrixFile):
        """
        def loadMatrix(matrixFile): loads a projection matrix from ASCII file
        """
        f = open( matrixFile, 'r')
        self.P = Matrix([1.0,0,0,0],[0,1.0,0,0],[0,0,1.0,0],[0,0,0,1.0])
        lines = f.readlines()   
        row = 0
        for line in lines:
            nums = re.findall( "(-*\d+\.*\d*)\s*,*;*", line)             
            for col in range(4):                   
                self.P[row][col] = float( nums[col])        
            row = row+1
        f.close()   
    def toTexture(self):
            """
            def toTexture(): convert image to a texture 
            and creates a material for that
            """            
            self.mat = Material.New()          # create a new Material called 'newMat'
            self.mat.setName(self.img.getName())
            #self.mat.
            self.mat.rgbCol = [1.0, 1.0, 1.0]          # change its color
            #mat.setAlpha(0.2)                     # mat.alpha = 0.2 -- almost transparent
            #mat.emit = 0.7                        # equivalent to mat.setEmit(0.8)
            #mat.mode |= Material.Modes.ZTRANSP    # turn on Z-Buffer transparency
            #mat.setName('RedBansheeSkin')         # change its name
            #mat.setAdd(0.8)                       # make it glow
            #mat.setMode('Halo')         
            #mat.setTexture( 0, texture, TexCo.ORCO, MapTo.COL)
            self.tex = Texture.New()             # get texture named 'foo'
            #tex = Texture.New()
            #print "testx"
            self.tex.setType("Image")                    # make foo be an image texture
            #img = Image.New( filename, 800, 600, 24)
            #img.Load(filename)              # load an image
            #print "test"
            #print "size",img.getSize()
            self.tex.setImage( self.img)                       # link the image to the texture
            #self.tex.d
            #mat = Material.Get('bar')               # get a material
            #mtextures = mat.getTextures()           # get a list of the MTex objects
            #for mtex in mtextures:
                #if mtex.tex.type == Texture.Types.IMAGE: 
                    #print mtex.tex.image.filename   # print the filenames of all the
                                           # images in textures linked to "bar"
            self.mat.setTexture(0, self.tex, Texture.TexCo.UV, Texture.MapTo.COL)               # set the material's first texture
                                           # to be our texture                                 
            self.mat.setMode("Shadeless")
            #tex = Texture.New()
            #print "oK"
    def getImage(self):
        """
        def getImage(self): returns associated blender image
        """
        return self.img
    def getPMatrix(self):
        """
        def getPMatrix(self): returns associated projection blender matrix
        Projections matrix is so defined that the [0,0] corner is top left
        """
        return self.P
    def getMaterial(self):
        """
        def getMaterial(self): returns associated blender material
        """
        return self.mat
    def getTexture(self):
        """
        def getTexture(self): returns associated blender texture
        """
        return self.tex
    def findBoundaries(self, background=[0,0,0,0]):
        """
        def findBoundaries(self, background=[0,0,0,0]): return list 
        of [left, bottom, right, top]
        boundary of the non-background colors in the image
        background color is defined as RGBA values 0-255
        """
        image = self.img
        minXY = image.getMinXY()
        maxXY = image.getMaxXY()
        ref = background
        left=maxXY[0]
        right=minXY[0]
        top=minXY[1]
        bottom=maxXY[1]
        for x in range(minXY[0],maxXY[0]):
            for y in range(minXY[1],maxXY[1]):
                if ref != image.getPixelI(x,y):
                    if top < y :
                        top = y
                    if bottom > y:
                        bottom = y
                    if left > x:
                        left = x
                    if right < x:
                        right = x
        return [left,bottom,right,top]
    def renderToImage( self,meshObject,meshColorIndex=1):
        """
        def renderToImage( meshObject, meshColorIndex=1): renders a Mesh Object
        into the image with the color index meshColorIndex. returns buffer with 
        RGBA values.
        """
        initRender()
        BGL.glViewport(0,0,self.getImage().size[0],self.getImage().size[1])
        BGL.glRasterPos3i(-1, -1, 1)        
        BGL.glDrawPixels(self.getImage().size[0],self.getImage().size[1], GL_RGBA, GL_UNSIGNED_BYTE, self.getBitmap())
        BGL.glRasterPos2i(0,0)
        setProjection(  self.getPMatrix(), 
                        self.getImage().size[0], 
                        self.getImage().size[1])
        renderObject( meshObject, meshColorIndex)                       
        self.bitmap = getRenderData(self.getImage().size[0],self.getImage().size[1], False)
        for y in range( self.img.size[1]):
            for x in range( self.img.size[0]):
                self.img.setPixelI( x, y, self.bitmap[y][x])

    def toBitmap( self):
        """
        def toBitmap( self): returns Buffer object with GL_BYTE values with image
        bitmap. Copy Image to bitmap.
        """
        if self.img is None:
            return None
        self.bitmap = Buffer( GL_BYTE, [self.img.size[1],self.img.size[0],4])
        for y in range( self.img.size[1]):
            for x in range( self.img.size[0]):
                RGBA = self.img.getPixelI( x, y)
                self.bitmap[y][x][0] = RGBA[0]
                self.bitmap[y][x][1] = RGBA[1]
                self.bitmap[y][x][2] = RGBA[2]
                #                self.bitmap[y][x][3] = RGBA[3]                
                if ((self.bitmap[y][x][0] == 0) and
                    (self.bitmap[y][x][1] == 0) and
                    (self.bitmap[y][x][2] == 0)):
                    self.bitmap[y][x][3] = 0
                else:
                    self.bitmap[y][x][3] = 255
        return self.bitmap
    def getBitmap( self):
        """
        def toBitmap( self): returns Buffer object with GL_BYTE values with image
        bitmap. Useful for OpenGL.
        """
        return self.bitmap
    def project( self, x, y, z):
        """
        def project( x, y, z):  return list ( u, v) coordinates in the picture
        using projective transformation.
        """
        # OLD VERSION
        v = Vector( [x, y, z, 1.0] )
        result = self.P * v
        # debug print
        # print self.P," * ",v,"=",result
        return [result[0]/result[2]/result[3], result[1]/result[2]/result[3]]
        # NEW - USING GLU PROJECT
        #glMatrixMode( GL_MODELVIEW)
        #glPushMatrix()
        #glMatrixMode( GL_PROJECTION)
        #glPushMatrix()
        #setProjection(  self.getPMatrix(), 
        #                self.getImage().size[0], 
        #                self.getImage().size[1]) # configure projections
        #glMatrixMode( GL_MODELVIEW)
        # Get the Projection matrix 4x4 
        #projection = Buffer( GL_DOUBLE, [16]) 
        #glGetDoublev( GL_PROJECTION_MATRIX, projection)
        # Get the viewport dimensions
        #view = Buffer( GL_INT, [4]) 
        #glGetIntegerv( GL_VIEWPORT, view)
        # Get the final model transformations        
        #model = Buffer( GL_DOUBLE, [16]) 
        #glGetDoublev( GL_MODELVIEW_MATRIX, model)  
        # We need a buffers for each value
        #UVlist = list()
        #wx = Buffer( GL_DOUBLE, 1)   # protoze gluProject vraci double
        #wy = Buffer( GL_DOUBLE, 1)   
        #wz = Buffer( GL_DOUBLE, 1)    
        #gluProject( x, y, z, model, projection, view, wx, wy, wz)  
        #glMatrixMode( GL_MODELVIEW)
        #glPopMatrix()
        #glMatrixMode( GL_PROJECTION)
        #glPopMatrix()
        #return Vector([wx,len(self.bitmap)-wy])
    def makeCopy( self):
        """
        def makeCopy( self): makes copy of itself
        """
        ImageObj = Image.New("copy", self.img.size[0], self.img.size[1], 24)
        for y in range( self.img.size[1]):
            for x in range( self.img.size[0]):
                ImageObj.setPixelI( x, y, self.img.getPixelI( x, y))
        Pmatrix = Matrix( self.P)
        c = CalibIm()
        c.img = ImageObj
        c.P = Pmatrix
        c.toBitmap()
        return c
    def getColIndexAt( self, u, v):
        """
        def getColIndexAt( u, v): returns color index at position (u,v)
        where the uv coordinates are follows:
        +-------> v
        |
        |
        V u
        but in blender image coordinates are:
        ^ y
        |
        |
        +----> x
        """
        if (int(v) < len(self.bitmap[0])) and (int(u) < len(self.bitmap)) and (int(u) >= 0) and (int(v) >= 0):
            return colToIndex( self.bitmap[len(self.bitmap)-int(u)][int(v)])
        else:
            return colToIndex( [0,0,0,0])
    def getFacePixels( self, object):
        """
        def getFacePixels( self, object): returns a list of 
        [ face id, visible pixels, hidden pixels] in this image
        object must a blender object with MeshObject Data
        """
        self.initRendering()        
        return getVisibleFaces( object)
    def getFaceUVs( self, object):
        """
        def getFaceUVs( object): returns a list of Vectors uv coordinates [ [[u1,v1],[u2,v2],..],..]
        for each face of the mesh. UV coordinates origin is the bottom left corner.
        """
        # check proper arguments
        if object.getType() != "Mesh" :        
            print "error! def getUV(object): given object is not a Mesh type"
            print "       object type is: " + object.getType()
            return None  
        # store current transformations settings
        glMatrixMode( GL_MODELVIEW)
        glPushMatrix()
        glMatrixMode( GL_PROJECTION)
        glPushMatrix()
        setProjection(  self.getPMatrix(), 
                        self.getImage().size[0], 
                        self.getImage().size[1]) # configure projections
        glMatrixMode( GL_MODELVIEW)
        glMultMatrixd( matToBuf( Matrix(object.getMatrix()).transpose())) # set model transformations
        # Get the Projection matrix 4x4 
        projection = Buffer( GL_DOUBLE, [16]) 
        glGetDoublev( GL_PROJECTION_MATRIX, projection)
        # Get the viewport dimensions
        view = Buffer( GL_INT, [4]) 
        glGetIntegerv( GL_VIEWPORT, view)
        # Get the final model transformations        
        model = Buffer( GL_DOUBLE, [16]) 
        glGetDoublev( GL_MODELVIEW_MATRIX, model)  
        # We need a buffers for each value
        UVlist = list()
        wx = Buffer( GL_DOUBLE, 1)   # protoze gluProject vraci double
        wy = Buffer( GL_DOUBLE, 1)   
        wz = Buffer( GL_DOUBLE, 1)    
        # Get the faces and uv coordinates for vertices
        nmesh = getDeformedMeshData(object)   
        faceid = 0
        for face in nmesh.faces:   
            uvs = list()     
            for ver in face.v:
                gluProject( ver[0], ver[1], ver[2], model, projection, view, wx, wy, wz)            
                uvs.append( Vector( wx[0]/self.getImage().size[0], wy[0]/self.getImage().size[1]) )
            UVlist.append( uvs)
            faceid = faceid + 1   
        # restore transformations settings
        glMatrixMode( GL_MODELVIEW)
        glPopMatrix()
        glMatrixMode( GL_PROJECTION)
        glPopMatrix()
        #print "UV list:", UVlist     
        return UVlist        
    def initRendering( self):
        """
        def initRendering(): calls initRender and setProjection with own projection matrix
        It cleans transformations and buffers and sets the own projection - can be obtained from getPMatrix()
        """
        initRender()
        setProjection(  self.getPMatrix(), 
                        self.getImage().size[0], 
                        self.getImage().size[1]) 
    def getUVFromView( self, vector):
        """
        def getUVFromView( vector): returns vector s uv coordinates in image using the gluProject
        in the current transformation settings.
        It doesn t change the current modelview or projection transformations.
        The origin of uv coordinates is in the bottom left corner, axes facing up and right.
        vector is the (x,y,z) Vector object or a list of floats
        returns (u,v)
        """
        projection = Buffer( GL_DOUBLE, [16]) 
        glGetDoublev( GL_PROJECTION_MATRIX, projection)
        view = Buffer( GL_INT, [4]) 
        glGetIntegerv( GL_VIEWPORT, view)
        model = Buffer( GL_DOUBLE, [16]) 
        glGetDoublev( GL_MODELVIEW_MATRIX, model)  
        wx = Buffer( GL_DOUBLE, 1)   
        wy = Buffer( GL_DOUBLE, 1)   
        wz = Buffer( GL_DOUBLE, 1)    
        gluProject( vector[0], vector[1], vector[2], model, projection, view, wx, wy, wz)            
        return [wx[0],wy[0]]